#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>

int readBytes(int sock, unsigned char *buf, int size);

int main (int argc, char *argv[])
{
    
    int listenfd, camera_sock = -1, viewer_sock = -1, connfd;
    int client_len, flag = 1;

    /* An array of Client sockets */
    int client[FD_SETSIZE];
    int i, nready;

    /* Sets of file descriptors for I/O multiplexing */
    fd_set rset, allset;

    /* Socket address structure instances for binding and accepting */
    struct sockaddr_in serv_addr, cliaddr;

    int maxfd, maxi, sockfd, port = 2112;


    if (argc != 2){
        printf("usage: %s port\n", argv[0]);
        exit(1);
    }
    port = atoi(argv[1]);



    /* Create listen socket */
    listenfd = socket(AF_INET,SOCK_STREAM,0);
    if (listenfd == -1)
    {
        perror("socket");
        exit(1);
    }

    /* Set SO_REUSEADDR option to reuse server binding address */
    if (setsockopt(listenfd,SOL_SOCKET,SO_REUSEADDR,(char *)&flag,sizeof(flag)) == -1)
    {
        perror("setsockopt");
        exit(1);
    }
    
    /* Initialize server address instance */  
    memset(&serv_addr,0, sizeof(serv_addr));
    serv_addr.sin_family = AF_INET;
    serv_addr.sin_addr.s_addr = inet_addr("0.0.0.0");
    serv_addr.sin_port = htons(port);

    /* Socket binding */
    if (bind(listenfd,(struct sockaddr *)&serv_addr,sizeof(serv_addr)) == -1){
        perror("bind");
        exit(1);
    }

    if (listen(listenfd,5) == -1){
        perror("listen");
        exit(1);
    }
    
    maxfd = listenfd; /* initialize */
    maxi = -1; /* index into client[] array */
    for (i = 0; i < FD_SETSIZE; i++){
        client[i] = -1; /* -1 indicates available entry */
    }

    FD_ZERO(&allset);
    FD_SET(listenfd, &allset);
    
    while(1){
        rset = allset; /* structure assignment */
        nready = select(maxfd + 1, &rset, NULL, NULL, NULL); 

        if (FD_ISSET(listenfd, &rset)){ /* new client connection */
            client_len = sizeof(cliaddr);
            connfd = accept(listenfd, (struct sockaddr *)&cliaddr, &client_len);

            for (i=0; i<FD_SETSIZE; i++){
                if (client[i] < 0){
                    client[i] = connfd; /* save descriptor */
                    break;
                }
            }

            if (i == FD_SETSIZE){ 
                printf("Too many peers");
                exit(1);
            }
            /* Set viewer socket and camera socket */
            if (viewer_sock == -1){
                viewer_sock = connfd;
                printf("viewer = %d\n", viewer_sock);
            }else if (camera_sock == -1){
                camera_sock = connfd;
                printf("camera = %d\n", camera_sock);
            }

            FD_SET(connfd, &allset); /* add new descriptor to set */
            if (connfd > maxfd)
                maxfd = connfd; /* for select */
            if (i > maxi)
                maxi = i; /* max index in client[] array */
            if (--nready <= 0)
                continue; /*  no more readable descriptors */
        }

        for (i=0; i<=maxi; i++){ /* check all clients for data */
            if ((sockfd = client[i]) < 0)
                continue;
            if (FD_ISSET(sockfd, &rset)){
                int size = 0, n;
                unsigned char *buf;
                if ( (n = readBytes(sockfd, &size, sizeof(size))) == 0){
                    /* connection closed by client */
                    close(sockfd);
                    FD_CLR(sockfd, &allset);
                    client[i] = -1;
                    if (sockfd == camera_sock){
                        camera_sock = -1;
                    }else if (sockfd == viewer_sock){
                        viewer_sock = -1;
                    }
                }else if (n == sizeof(int)){
                    /* Read more data from a camera or a camera viewer socket*/
                    size = ntohl(size);
                    printf("received packet size = %d\n", size);
                    buf = malloc(size);
                    readBytes(sockfd, buf, size);
                    if (sockfd == camera_sock){
                        /* Send image data from a camera to a camera viewer */
                        int tmp = ntohl(size);
                        printf("camera -> viewer\n");
                        write(viewer_sock, &tmp, sizeof(tmp));
                        write(viewer_sock, buf, size);
                    }else if(sockfd == viewer_sock){
                        /* Send a control signal from a camera viewer to a camera */
                        int tmp = ntohl(size);
                        printf("viewer -> camera\n");
                        write(camera_sock, &tmp, sizeof(tmp));
                        write(camera_sock, buf, size);
                    }
                    free(buf);
                }

                if (--nready <= 0){
                    break; /* no more readable descriptors */
                }
            }
        }
    }
    return 0;
}
/*
 * readBytes() attempts to read up to size bytes from file descriptor fd 
 * into the buffer starting at buf.
 *
 * If count is zero, readBytes return zero and has no other results. 
 *
 * On success, the number of bytes read is returned.
 * On error, -1 is returned.
 */
int readBytes(int sock, unsigned char *buf, int size){
    int read_count = 0;
    if (size < 0){
        return -1;
    }

    while(size > 0){
        int ret;
        ret = read(sock, buf, size);
        if (ret > 0){
            size -= ret;
            read_count += ret;
            buf += ret;
        }else if(ret == 0){
            return read_count;
        }else if(ret == -1){
            return -1;
        }
    }
    return read_count;
}



